@@ -29,10 +29,15 @@ struct libcrun_context_s
   const char *id;
   const char *bundle;
   const char *config_file;
+  const char *config_file_content;
   const char *console_socket;
   const char *pid_file;
   const char *notify_socket;
+  const char *handler;
   int preserve_fds;
+  // For some use-cases we need differentiation between preserve_fds and listen_fds.
+  // Following context variable makes sure we get exact value of listen_fds irrespective of preserve_fds.
+  int listen_fds;
 
   crun_output_handler output_handler;
   void *output_handler_arg;
@@ -69,6 +74,7 @@ struct libcrun_container_s
   bool use_intermediate_userns;
 
   void *private_data;
+  void (*cleanup_private_data) (void *private_data);
   struct libcrun_context_s *context;
 };
 
@@ -87,6 +93,7 @@ struct libcrun_checkpoint_restore_s
   bool shell_job;
   bool ext_unix_sk;
   bool detach;
+  bool file_locks;
   const char *console_socket;
 };
 typedef struct libcrun_checkpoint_restore_s libcrun_checkpoint_restore_t;
@@ -95,6 +102,8 @@ LIBCRUN_PUBLIC libcrun_container_t *libcrun_container_load_from_file (const char
 
 LIBCRUN_PUBLIC libcrun_container_t *libcrun_container_load_from_memory (const char *json, libcrun_error_t *err);
 
+LIBCRUN_PUBLIC void libcrun_container_free (libcrun_container_t *);
+
 LIBCRUN_PUBLIC int libcrun_container_run (libcrun_context_t *context, libcrun_container_t *container,
                                           unsigned int options, libcrun_error_t *error);
 
@@ -119,6 +128,18 @@ LIBCRUN_PUBLIC int libcrun_get_container_state_string (const char *id, libcrun_c
                                                        const char *state_root, const char **container_status,
                                                        int *running, libcrun_error_t *err);
 
+struct libcrun_container_exec_options_s
+{
+  size_t struct_size;
+  runtime_spec_schema_config_schema_process *process;
+  const char *path;
+  const char *cgroup;
+};
+
+LIBCRUN_PUBLIC int libcrun_container_exec_with_options (libcrun_context_t *context, const char *id,
+                                                        struct libcrun_container_exec_options_s *opts,
+                                                        libcrun_error_t *err);
+
 LIBCRUN_PUBLIC int libcrun_container_exec (libcrun_context_t *context, const char *id,
                                            runtime_spec_schema_config_schema_process *process, libcrun_error_t *err);
 
@@ -143,4 +164,17 @@ LIBCRUN_PUBLIC int libcrun_container_checkpoint (libcrun_context_t *context, con
 LIBCRUN_PUBLIC int libcrun_container_restore (libcrun_context_t *context, const char *id,
                                               libcrun_checkpoint_restore_t *cr_options, libcrun_error_t *err);
 
+// Not part of the public API, just a method in container.c we need to access from linux.c
+void get_root_in_the_userns (runtime_spec_schema_config_schema *def, uid_t host_uid, gid_t host_gid,
+                             uid_t *uid, gid_t *gid);
+
+static inline void
+cleanup_containerp (libcrun_container_t **c)
+{
+  libcrun_container_t *container = *c;
+  libcrun_container_free (container);
+}
+
+#define cleanup_container __attribute__ ((cleanup (cleanup_containerp)))
+
 #endif
