@@ -39,10 +39,17 @@
 #  define CRIU_RESTORE_LOG_FILE "restore.log"
 #  define DESCRIPTORS_FILENAME "descriptors.json"
 
+#  define CRIU_EXT_NETNS "extRootNetNS"
+#  define CRIU_EXT_PIDNS "extRootPidNS"
+
+#  ifndef CLONE_NEWTIME
+#    define CLONE_NEWTIME 0x00000080 /* New time namespace */
+#  endif
+
 static const char *console_socket = NULL;
 
 static int
-criu_notify (char *action, criu_notify_arg_t na)
+criu_notify (char *action, __attribute__ ((unused)) criu_notify_arg_t na)
 {
   if (strncmp (action, "orphan-pts-master", 17) == 0)
     {
@@ -76,10 +83,10 @@ restore_cgroup_v1_mount (runtime_spec_schema_config_schema *def, libcrun_error_t
   int cgroup_mode;
   char *from;
   int ret;
-  int i;
+  uint32_t i;
 
   cgroup_mode = libcrun_get_cgroup_mode (err);
-  if (cgroup_mode < 0)
+  if (UNLIKELY (cgroup_mode < 0))
     return cgroup_mode;
 
   if (cgroup_mode == CGROUP_MODE_UNIFIED)
@@ -151,7 +158,7 @@ checkpoint_cgroup_v1_mount (runtime_spec_schema_config_schema *def, libcrun_erro
   char *saveptr = NULL;
   char *from;
   int ret;
-  int i;
+  uint32_t i;
 
   /* First check if there is actually a cgroup mount in the container. */
   for (i = 0; i < def->mounts_len; i++)
@@ -213,8 +220,6 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
   runtime_spec_schema_config_schema *def = container->container_def;
   cleanup_free char *descriptors_path = NULL;
   cleanup_free char *freezer_path = NULL;
-  cleanup_close int descriptors_fd = -1;
-  cleanup_free char *external = NULL;
   cleanup_free char *path = NULL;
   cleanup_close int image_fd = -1;
   cleanup_close int work_fd = -1;
@@ -244,18 +249,18 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
 
   ret = criu_init_opts ();
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, 0, "CRIU init failed with %d\n", ret);
+    return crun_make_error (err, 0, "CRIU init failed with %d", ret);
 
   if (UNLIKELY (cr_options->image_path == NULL))
-    return crun_make_error (err, 0, "image path not set\n");
+    return crun_make_error (err, 0, "image path not set");
 
   ret = mkdir (cr_options->image_path, 0700);
   if (UNLIKELY ((ret == -1) && (errno != EEXIST)))
-    return crun_make_error (err, errno, "error creating checkpoint directory %s\n", cr_options->image_path);
+    return crun_make_error (err, errno, "error creating checkpoint directory %s", cr_options->image_path);
 
   image_fd = open (cr_options->image_path, O_DIRECTORY);
   if (UNLIKELY (image_fd == -1))
-    return crun_make_error (err, errno, "error opening checkpoint directory %s\n", cr_options->image_path);
+    return crun_make_error (err, errno, "error opening checkpoint directory %s", cr_options->image_path);
 
   criu_set_images_dir_fd (image_fd);
 
@@ -265,16 +270,9 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
   if (UNLIKELY (ret < 0))
     return ret;
 
-  descriptors_fd = open (descriptors_path, O_CREAT | O_WRONLY | O_CLOEXEC, S_IRUSR | S_IWUSR);
-  if (UNLIKELY (descriptors_fd == -1))
-    return crun_make_error (err, errno, "error opening descriptors file %s\n", descriptors_path);
-  if (status->external_descriptors)
-    {
-      ret = TEMP_FAILURE_RETRY (
-          write (descriptors_fd, status->external_descriptors, strlen (status->external_descriptors)));
-      if (UNLIKELY (ret < 0))
-        return crun_make_error (err, errno, "write '%s'", DESCRIPTORS_FILENAME);
-    }
+  ret = write_file (descriptors_path, status->external_descriptors, strlen (status->external_descriptors), err);
+  if (UNLIKELY (ret < 0))
+    return crun_error_wrap (err, "error saving CRIU descriptors file");
 
   /* work_dir is the place CRIU will put its logfiles. If not explicitly set,
    * CRIU will put the logfiles into the images_dir from above. No need for
@@ -283,7 +281,7 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
     {
       work_fd = open (cr_options->work_path, O_DIRECTORY);
       if (UNLIKELY (work_fd == -1))
-        return crun_make_error (err, errno, "error opening CRIU work directory %s\n", cr_options->work_path);
+        return crun_make_error (err, errno, "error opening CRIU work directory %s", cr_options->work_path);
 
       criu_set_work_dir_fd (work_fd);
     }
@@ -303,10 +301,10 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
 
   ret = criu_set_root (path);
   if (UNLIKELY (ret != 0))
-    return crun_make_error (err, 0, "error setting CRIU root to %s\n", path);
+    return crun_make_error (err, 0, "error setting CRIU root to %s", path);
 
   cgroup_mode = libcrun_get_cgroup_mode (err);
-  if (cgroup_mode < 0)
+  if (UNLIKELY (cgroup_mode < 0))
     return cgroup_mode;
 
   /* For cgroup v1 we need to tell CRIU to handle all cgroup mounts as external mounts. */
@@ -314,7 +312,7 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
     {
       ret = checkpoint_cgroup_v1_mount (def, err);
       if (UNLIKELY (ret != 0))
-        return crun_make_error (err, 0, "error handling cgroup v1 mounts\n");
+        return crun_make_error (err, 0, "error handling cgroup v1 mounts");
     }
 
   /* Tell CRIU about external bind mounts. */
@@ -340,22 +338,25 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
         criu_add_ext_mount (def->linux->masked_paths[i], def->linux->masked_paths[i]);
     }
 
-  /* CRIU tries to checkpoint and restore all namespaces. The network
-   * namespace, however, is usually a bit special as it requires
-   * connecting network interfaces into the namespace. CRIU has support
-   * to reconnect veth devices, but as we are mainly targeting Podman
-   * right now, only Podman's way of creating network namespaces via
-   * CNI is handled here. We are looking at config.json and if there
-   * is a path configured for the network namespace we are telling CRIU
-   * to ignore the network namespace and just restore the container
-   * into the existing network namespace.
+  /* CRIU tries to checkpoint and restore all namespaces. However,
+   * namespaces could be shared between containers in a pod.
+   * To address this, CRIU provides support for external namespaces.
+   * External namespaces allow to ignore the namespace during checkpoint
+   * and restore the container into the existing namespaces.
+   *
+   * We are looking at config.json and if there is a path configured for
+   * a namespace we are telling CRIU to ignore the namespace and
+   * just restore the container into the existing namespace.
+   *
+   * In the case of Podman, a network namespace would be created via CNI.
    *
    * CRIU expects the information about an external namespace like this:
-   * --external net[<inode>]:<key>
-   * For key we are using the string: 'extRootNetNS'. */
+   * --external <namespace>[<inode>]:<key>
+   */
 
   for (i = 0; i < def->linux->namespaces_len; i++)
     {
+      cleanup_free char *external = NULL;
       int value = libcrun_find_namespace (def->linux->namespaces[i]->type);
       if (UNLIKELY (value < 0))
         return crun_make_error (err, 0, "invalid namespace type: `%s`", def->linux->namespaces[i]->type);
@@ -367,9 +368,19 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
           if (UNLIKELY (ret < 0))
             return crun_make_error (err, errno, "unable to stat(): `%s`", def->linux->namespaces[i]->path);
 
-          xasprintf (&external, "net[%ld]:extRootNetNS", statbuf.st_ino);
+          xasprintf (&external, "net[%ld]:" CRIU_EXT_NETNS, statbuf.st_ino);
+          criu_add_external (external);
+        }
+
+      if (value == CLONE_NEWPID && def->linux->namespaces[i]->path != NULL)
+        {
+          struct stat statbuf;
+          ret = stat (def->linux->namespaces[i]->path, &statbuf);
+          if (UNLIKELY (ret < 0))
+            return crun_make_error (err, errno, "unable to stat(): `%s`", def->linux->namespaces[i]->path);
+
+          xasprintf (&external, "pid[%ld]:" CRIU_EXT_PIDNS, statbuf.st_ino);
           criu_add_external (external);
-          break;
         }
     }
 
@@ -390,13 +401,14 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
 
   ret = criu_set_freeze_cgroup (freezer_path);
   if (UNLIKELY (ret != 0))
-    return crun_make_error (err, ret, "CRIU: failed setting freezer %d\n", ret);
+    return crun_make_error (err, ret, "CRIU: failed setting freezer %d", ret);
 
   /* Set boolean options . */
   criu_set_leave_running (cr_options->leave_running);
   criu_set_ext_unix_sk (cr_options->ext_unix_sk);
   criu_set_shell_job (cr_options->shell_job);
   criu_set_tcp_established (cr_options->tcp_established);
+  criu_set_file_locks (cr_options->file_locks);
   criu_set_orphan_pts_master (true);
   criu_set_manage_cgroups (true);
 
@@ -406,8 +418,7 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
   ret = criu_dump ();
   if (UNLIKELY (ret != 0))
     return crun_make_error (err, 0,
-                            "CRIU checkpointing failed %d\n"
-                            "Please check CRIU logfile %s/%s\n",
+                            "CRIU checkpointing failed %d.  Please check CRIU logfile %s/%s",
                             ret, cr_options->work_path, CRIU_CHECKPOINT_LOG_FILE);
 
   return 0;
@@ -416,7 +427,7 @@ libcrun_container_checkpoint_linux_criu (libcrun_container_status_t *status, lib
 static int
 prepare_restore_mounts (runtime_spec_schema_config_schema *def, char *root, libcrun_error_t *err)
 {
-  int i;
+  uint32_t i;
 
   /* Go through all mountpoints to be able to recreate missing mountpoints. */
   for (i = 0; i < def->mounts_len; i++)
@@ -493,7 +504,8 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
                                       libcrun_checkpoint_restore_t *cr_options, libcrun_error_t *err)
 {
   runtime_spec_schema_config_schema *def = container->container_def;
-  cleanup_close int inherit_fd = -1;
+  cleanup_close int inherit_new_net_fd = -1;
+  cleanup_close int inherit_new_pid_fd = -1;
   cleanup_close int image_fd = -1;
   cleanup_free char *root = NULL;
   cleanup_close int work_fd = -1;
@@ -506,14 +518,14 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
 
   ret = criu_init_opts ();
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, 0, "CRIU init failed with %d\n", ret);
+    return crun_make_error (err, 0, "CRIU init failed with %d", ret);
 
   if (UNLIKELY (cr_options->image_path == NULL))
-    return crun_make_error (err, 0, "image path not set\n");
+    return crun_make_error (err, 0, "image path not set");
 
   image_fd = open (cr_options->image_path, O_DIRECTORY);
   if (UNLIKELY (image_fd == -1))
-    return crun_make_error (err, errno, "error opening checkpoint directory %s\n", cr_options->image_path);
+    return crun_make_error (err, errno, "error opening checkpoint directory %s", cr_options->image_path);
 
   criu_set_images_dir_fd (image_fd);
 
@@ -532,8 +544,6 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
     if (UNLIKELY (ret < 0))
       return ret;
 
-    status->external_descriptors = xstrdup (buffer);
-
     /* descriptors.json contains a JSON array with strings
      * telling where 0, 1 and 2 have been initially been
      * pointing to. For each descriptor which points to
@@ -556,7 +566,7 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
             if (s && YAJL_IS_STRING (s))
               {
                 char *str = YAJL_GET_STRING (s);
-                if (strncmp (str, "pipe:", 5) == 0)
+                if (has_prefix (str, "pipe:"))
                   criu_add_inherit_fd (i, str);
               }
           }
@@ -571,7 +581,7 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
     {
       work_fd = open (cr_options->work_path, O_DIRECTORY);
       if (UNLIKELY (work_fd == -1))
-        return crun_make_error (err, errno, "error opening CRIU work directory %s\n", cr_options->work_path);
+        return crun_make_error (err, errno, "error opening CRIU work directory %s", cr_options->work_path);
 
       criu_set_work_dir_fd (work_fd);
     }
@@ -611,12 +621,12 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
 
   ret = mkdir (root, 0755);
   if (UNLIKELY (ret == -1))
-    return crun_make_error (err, errno, "error creating restore directory %s\n", root);
+    return crun_make_error (err, errno, "error creating restore directory %s", root);
   /* do realpath on root */
   ret = mount (status->rootfs, root, NULL, MS_BIND | MS_REC, NULL);
   if (UNLIKELY (ret == -1))
     {
-      ret = crun_make_error (err, errno, "error mounting restore directory %s\n", root);
+      ret = crun_make_error (err, errno, "error mounting restore directory %s", root);
       goto out;
     }
 
@@ -634,13 +644,23 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
   ret = criu_set_root (root);
   if (UNLIKELY (ret != 0))
     {
-      ret = crun_make_error (err, 0, "error setting CRIU root to %s\n", root);
+      ret = crun_make_error (err, 0, "error setting CRIU root to %s", root);
       goto out_umount;
     }
 
-  /* If there is network namespace defined in config.json we are telling
-   * CRIU to restore the process into that network namespace.
-   * CRIU expects the information about the network namespace like this:
+#  ifdef CRIU_JOIN_NS_SUPPORT
+  /* criu_join_ns_add() API was introduced with CRIU version 3.16.1
+   * Here we check if this API is available at build time to support
+   * compiling with older version of CRIU, and at runtime to support
+   * running crun with older versions of libcriu.so.2.
+   */
+  bool join_ns_support = criu_check_version (31601) == 1;
+#  endif
+
+  /* If a namespace defined in config.json we are telling
+   * CRIU use that namespace when restoring the process tree.
+   *
+   * CRIU expects the information about the namespace like this:
    * --inherit-fd fd[<fd>]:<key>
    * The <key> needs to be the same as during checkpointing (extRootNetNS). */
   for (i = 0; i < def->linux->namespaces_len; i++)
@@ -651,13 +671,47 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
 
       if (value == CLONE_NEWNET && def->linux->namespaces[i]->path != NULL)
         {
-          inherit_fd = open (def->linux->namespaces[i]->path, O_RDONLY);
-          if (UNLIKELY (ret < 0))
+          inherit_new_net_fd = open (def->linux->namespaces[i]->path, O_RDONLY);
+          if (UNLIKELY (inherit_new_net_fd < 0))
             return crun_make_error (err, errno, "unable to open(): `%s`", def->linux->namespaces[i]->path);
 
-          criu_add_inherit_fd (inherit_fd, "extRootNetNS");
-          break;
+          criu_add_inherit_fd (inherit_new_net_fd, CRIU_EXT_NETNS);
         }
+
+      if (value == CLONE_NEWPID && def->linux->namespaces[i]->path != NULL)
+        {
+          inherit_new_pid_fd = open (def->linux->namespaces[i]->path, O_RDONLY);
+          if (UNLIKELY (inherit_new_pid_fd < 0))
+            return crun_make_error (err, errno, "unable to open(): `%s`", def->linux->namespaces[i]->path);
+
+          criu_add_inherit_fd (inherit_new_pid_fd, CRIU_EXT_PIDNS);
+        }
+
+#  ifdef CRIU_JOIN_NS_SUPPORT
+      if (value == CLONE_NEWTIME && def->linux->namespaces[i]->path != NULL)
+        {
+          if (join_ns_support)
+            criu_join_ns_add ("time", def->linux->namespaces[i]->path, NULL);
+          else
+            return crun_make_error (err, 0, "Shared time namespace restore is supported in CRIU >= 3.16.1");
+        }
+
+      if (value == CLONE_NEWIPC && def->linux->namespaces[i]->path != NULL)
+        {
+          if (join_ns_support)
+            criu_join_ns_add ("ipc", def->linux->namespaces[i]->path, NULL);
+          else
+            return crun_make_error (err, 0, "Shared ipc namespace restore is supported in CRIU >= 3.16.1");
+        }
+
+      if (value == CLONE_NEWUTS && def->linux->namespaces[i]->path != NULL)
+        {
+          if (join_ns_support)
+            criu_join_ns_add ("uts", def->linux->namespaces[i]->path, NULL);
+          else
+            return crun_make_error (err, 0, "Shared uts namespace restore is supported in CRIU >= 3.16.1");
+        }
+#  endif
     }
 
   /* Tell CRIU if cgroup v1 needs to be handled. */
@@ -672,6 +726,7 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
   criu_set_ext_unix_sk (cr_options->ext_unix_sk);
   criu_set_shell_job (cr_options->shell_job);
   criu_set_tcp_established (cr_options->tcp_established);
+  criu_set_file_locks (cr_options->file_locks);
   criu_set_orphan_pts_master (true);
   criu_set_manage_cgroups (true);
 
@@ -686,8 +741,7 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
   if (UNLIKELY (ret <= 0))
     {
       ret = crun_make_error (err, 0,
-                             "CRIU restoring failed %d\n"
-                             "Please check CRIU logfile %s/%s\n",
+                             "CRIU restoring failed %d.  Please check CRIU logfile %s/%s",
                              ret, cr_options->work_path, CRIU_RESTORE_LOG_FILE);
       goto out_umount;
     }
@@ -696,16 +750,22 @@ libcrun_container_restore_linux_criu (libcrun_container_status_t *status, libcru
    * be necessary later when moving the process into its cgroup. */
   status->pid = ret;
 
+  if (LIKELY (ret > 0))
+    ret = libcrun_save_external_descriptors (container, ret, err);
+
 out_umount:
   ret_out = umount (root);
   if (UNLIKELY (ret_out == -1))
-    return crun_make_error (err, errno, "error unmounting restore directory %s\n", root);
+    {
+      rmdir (root);
+      return crun_make_error (err, errno, "error unmounting restore directory %s", root);
+    }
 out:
   ret_out = rmdir (root);
   if (UNLIKELY (ret == -1))
     return ret;
   if (UNLIKELY (ret_out == -1))
-    return crun_make_error (err, errno, "error removing restore directory %s\n", root);
+    return crun_make_error (err, errno, "error removing restore directory %s", root);
   return ret;
 }
 #endif
