@@ -41,9 +41,12 @@
 #include <linux/magic.h>
 #include <limits.h>
 #include <stdarg.h>
+#ifdef HAVE_LINUX_OPENAT2_H
+#  include <linux/openat2.h>
+#endif
 
 #ifndef CLOSE_RANGE_CLOEXEC
-# define CLOSE_RANGE_CLOEXEC (1U << 2)
+#  define CLOSE_RANGE_CLOEXEC (1U << 2)
 #endif
 #ifndef RESOLVE_IN_ROOT
 #  define RESOLVE_IN_ROOT 0x10
@@ -108,7 +111,7 @@ xasprintf (char **str, const char *fmt, ...)
 int
 write_file_at (int dirfd, const char *name, const void *data, size_t len, libcrun_error_t *err)
 {
-  cleanup_close int fd = openat (dirfd, name, O_WRONLY | O_CREAT, 0700);
+  cleanup_close int fd = openat (dirfd, name, O_WRONLY | O_CREAT | O_TRUNC, 0700);
   int ret = 0;
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "opening file `%s` for writing", name);
@@ -141,7 +144,7 @@ write_file_with_flags (const char *name, int flags, const void *data, size_t len
 int
 write_file (const char *name, const void *data, size_t len, libcrun_error_t *err)
 {
-  return write_file_with_flags (name, O_CREAT, data, len, err);
+  return write_file_with_flags (name, O_CREAT | O_TRUNC, data, len, err);
 }
 
 int
@@ -165,7 +168,9 @@ get_file_type_fd (int fd, mode_t *mode)
   int ret;
 
 #ifdef HAVE_STATX
-  struct statx stx;
+  struct statx stx = {
+    0,
+  };
 
   ret = statx (fd, "", AT_EMPTY_PATH | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
   if (UNLIKELY (ret < 0))
@@ -192,7 +197,9 @@ get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
   int ret;
 
 #ifdef HAVE_STATX
-  struct statx stx;
+  struct statx stx = {
+    0,
+  };
 
   ret = statx (dirfd, path, (nofollow ? AT_SYMLINK_NOFOLLOW : 0) | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
   if (UNLIKELY (ret < 0))
@@ -329,38 +336,30 @@ close_and_replace (int *oldfd, int newfd)
 /* Defined in chroot_realpath.c  */
 char *chroot_realpath (const char *chroot, const char *path, char resolved_path[]);
 
-int
-safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path, int flags, int mode,
-             libcrun_error_t *err)
+static int
+safe_openat_fallback (int dirfd, const char *rootfs, size_t rootfs_len, const char *path,
+                      int flags, int mode, libcrun_error_t *err)
 {
-  int ret;
+  const char *path_in_chroot;
   cleanup_close int fd = -1;
-  static bool openat2_supported = true;
-  char buffer[PATH_MAX], *path_in_chroot;
+  char buffer[PATH_MAX];
+  int ret;
 
-  if (openat2_supported)
-    {
-      ret = syscall_openat2 (dirfd, path, flags, mode, RESOLVE_IN_ROOT);
-      if (ret < 0)
-        {
-          if (errno == ENOSYS)
-            openat2_supported = false;
-          if (errno == ENOSYS || errno == EINVAL || errno == EPERM)
-            goto fallback;
-          return crun_make_error (err, errno, "openat2 `%s`", path);
-        }
-
-      return ret;
-    }
-
-fallback:
   path_in_chroot = chroot_realpath (rootfs, path, buffer);
   if (path_in_chroot == NULL)
     return crun_make_error (err, errno, "cannot resolve `%s` under rootfs", path);
 
   path_in_chroot += rootfs_len;
-  while (*path_in_chroot == '/')
-    path_in_chroot++;
+  path_in_chroot = consume_slashes (path_in_chroot);
+
+  /* If the path is empty we are at the root, dup the dirfd itself.  */
+  if (path_in_chroot[0] == '\0')
+    {
+      ret = dup (dirfd);
+      if (UNLIKELY (ret < 0))
+        return crun_make_error (err, errno, "dup `%s`", rootfs);
+      return ret;
+    }
 
   ret = openat (dirfd, path_in_chroot, flags, mode);
   if (UNLIKELY (ret < 0))
@@ -377,8 +376,65 @@ fallback:
   return ret;
 }
 
+int
+safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path, int flags, int mode,
+             libcrun_error_t *err)
+{
+  static bool openat2_supported = true;
+  int ret;
+
+  if (openat2_supported)
+    {
+      ret = TEMP_FAILURE_RETRY (syscall_openat2 (dirfd, path, flags, mode, RESOLVE_IN_ROOT));
+      if (ret < 0)
+        {
+          if (errno == ENOSYS)
+            openat2_supported = false;
+          if (errno == ENOSYS || errno == EINVAL || errno == EPERM)
+            return safe_openat_fallback (dirfd, rootfs, rootfs_len, path, flags, mode, err);
+
+          return crun_make_error (err, errno, "openat2 `%s`", path);
+        }
+
+      return ret;
+    }
+
+  return safe_openat_fallback (dirfd, rootfs, rootfs_len, path, flags, mode, err);
+}
+
+static ssize_t
+safe_readlinkat (int dfd, const char *name, char **buffer, ssize_t hint, libcrun_error_t *err)
+{
+  ssize_t buf_size = hint > 0 ? hint + 1 : 512;
+  cleanup_free char *tmp_buf = NULL;
+  ssize_t size;
+
+  do
+    {
+      if (tmp_buf != NULL)
+        buf_size += 256;
+
+      /* Allocate an extra byte so the buffer can be NUL terminated.  */
+      tmp_buf = xrealloc (tmp_buf, buf_size + 1);
+
+      size = readlinkat (dfd, name, tmp_buf, buf_size);
+      if (UNLIKELY (size < 0))
+        return crun_make_error (err, errno, "readlink `%s`", name);
+  } while (size == buf_size);
+
+  /* Always NUL terminate the buffer.  */
+  tmp_buf[size] = '\0';
+
+  /* Move ownership to BUFFER.  */
+  *buffer = tmp_buf;
+  tmp_buf = NULL;
+
+  return size;
+}
+
 static int
-crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
+crun_safe_ensure_at (bool do_open, bool dir, int dirfd, const char *dirpath,
+                     size_t dirpath_len, const char *path, int mode,
                      int max_readlinks, libcrun_error_t *err)
 {
   cleanup_close int wd_cleanup = -1;
@@ -393,8 +449,7 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
   if (max_readlinks <= 0)
     return crun_make_error (err, ELOOP, "resolve path `%s`", path);
 
-  while (*path == '/')
-    path++;
+  path = consume_slashes (path);
 
   npath = xstrdup (path);
 
@@ -427,14 +482,14 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
             depth--;
           else
             {
+              /* Start from the root.  */
               close_and_reset (&wd_cleanup);
               cwd = dirfd;
+              goto next;
             }
         }
 
-      if (! last_component || dir)
-        ret = mkdirat (cwd, cur, mode);
-      else
+      if (last_component && ! dir)
         {
           ret = openat (cwd, cur, O_CLOEXEC | O_CREAT | O_WRONLY | O_NOFOLLOW, 0700);
           if (UNLIKELY (ret < 0))
@@ -442,25 +497,17 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
               /* If the last component is a symlink, repeat the lookup with the resolved path.  */
               if (errno == ELOOP)
                 {
-                  size_t s, resolved_size = 0;
                   cleanup_free char *resolved_path = NULL;
 
-                  do
+                  ret = safe_readlinkat (cwd, cur, &resolved_path, 0, err);
+                  if (LIKELY (ret >= 0))
                     {
-                      resolved_size += 512;
-                      resolved_path = xrealloc (resolved_path, resolved_size);
-
-                      s = readlinkat (dirfd, npath, resolved_path, resolved_size);
-                    }
-                  while (s == resolved_size);
-
-                  if (s > 0)
-                    {
-                      resolved_path[s] = '\0';
-                      crun_error_release (err);
-                      return crun_safe_ensure_at (dir, dirfd, dirpath, dirpath_len, resolved_path, mode,
+                      return crun_safe_ensure_at (do_open, dir, dirfd,
+                                                  dirpath, dirpath_len,
+                                                  resolved_path, mode,
                                                   max_readlinks - 1, err);
                     }
+                  crun_error_release (err);
                 }
               /* If the previous openat fails, attempt to open the file in O_PATH mode.  */
               ret = openat (cwd, cur, O_CLOEXEC | O_PATH, 0);
@@ -468,11 +515,14 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
                 return crun_make_error (err, errno, "open `%s/%s`", dirpath, cur);
             }
 
-          close_and_replace (&wd_cleanup, ret);
+          if (do_open)
+            return ret;
 
+          close_and_replace (&wd_cleanup, ret);
           return 0;
         }
 
+      ret = mkdirat (cwd, cur, mode);
       if (ret < 0)
         {
           if (errno != EEXIST)
@@ -489,26 +539,49 @@ crun_safe_ensure_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_le
       if (it == NULL)
         break;
 
-      cur = it + 1;
+      cur = consume_slashes (it + 1);
       *it = '/';
       it = strchr (cur, '/');
     }
 
+  if (do_open)
+    {
+      if (cwd == dirfd)
+        return dup (dirfd);
+
+      wd_cleanup = -1;
+      return cwd;
+    }
+
   return 0;
 }
 
+int
+crun_safe_create_and_open_ref_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_len,
+                                  const char *path, int mode, libcrun_error_t *err)
+{
+  int fd;
+
+  /* If the file/dir already exists, just open it.  */
+  fd = safe_openat (dirfd, dirpath, dirpath_len, path, O_PATH | O_CLOEXEC, 0, err);
+  if (LIKELY (fd >= 0))
+    return fd;
+
+  return crun_safe_ensure_at (true, dir, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
+}
+
 int
 crun_safe_ensure_directory_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
                                libcrun_error_t *err)
 {
-  return crun_safe_ensure_at (true, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
+  return crun_safe_ensure_at (false, true, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
 }
 
 int
 crun_safe_ensure_file_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
                           libcrun_error_t *err)
 {
-  return crun_safe_ensure_at (false, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
+  return crun_safe_ensure_at (false, false, dirfd, dirpath, dirpath_len, path, mode, MAX_READLINKS, err);
 }
 
 int
@@ -552,7 +625,9 @@ get_file_size (int fd, off_t *size)
   struct stat st;
   int ret;
 #ifdef HAVE_STATX
-  struct statx stx;
+  struct statx stx = {
+    0,
+  };
 
   ret = statx (fd, "", AT_EMPTY_PATH | AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC, STATX_SIZE, &stx);
   if (UNLIKELY (ret < 0))
@@ -595,12 +670,21 @@ int
 check_running_in_user_namespace (libcrun_error_t *err)
 {
   cleanup_free char *buffer = NULL;
+  static int run_in_userns = -1;
   size_t len;
-  int ret = read_all_file ("/proc/self/uid_map", &buffer, &len, err);
+  int ret;
+
+  ret = run_in_userns;
+  if (ret >= 0)
+    return ret;
+
+  ret = read_all_file ("/proc/self/uid_map", &buffer, &len, err);
   if (UNLIKELY (ret < 0))
     return ret;
 
-  return strstr (buffer, "4294967295") ? 0 : 1;
+  ret = strstr (buffer, "4294967295") ? 0 : 1;
+  run_in_userns = ret;
+  return ret;
 }
 
 static int selinux_enabled = -1;
@@ -824,12 +908,24 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
 {
   struct sockaddr_un addr = {};
   int ret;
-  cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
+  char name_buf[32];
+  cleanup_close int destfd = -1;
+  cleanup_close int fd = -1;
+
+  fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error creating UNIX socket");
 
   if (strlen (path) >= sizeof (addr.sun_path))
-    return crun_make_error (err, 0, "invalid path %s specified", path);
+    {
+      destfd = open (path, O_PATH);
+      if (UNLIKELY (destfd < 0))
+        return crun_make_error (err, errno, "error opening `%s`", path);
+
+      sprintf (name_buf, "/proc/self/fd/%d", destfd);
+      path = name_buf;
+    }
+
   strcpy (addr.sun_path, path);
   addr.sun_family = AF_UNIX;
   ret = connect (fd, (struct sockaddr *) &addr, sizeof (addr));
@@ -847,12 +943,16 @@ open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
 {
   struct sockaddr_un addr = {};
   int ret;
+  char name_buf[32];
   cleanup_close int fd = socket (AF_UNIX, dgram ? SOCK_DGRAM : SOCK_STREAM, 0);
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error creating UNIX socket");
 
   if (strlen (path) >= sizeof (addr.sun_path))
-    return crun_make_error (err, 0, "invalid path %s specified", path);
+    {
+      sprintf (name_buf, "/proc/self/fd/%d", fd);
+      path = name_buf;
+    }
   strcpy (addr.sun_path, path);
   addr.sun_family = AF_UNIX;
   ret = bind (fd, (struct sockaddr *) &addr, sizeof (addr));
@@ -874,18 +974,30 @@ open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err)
 
 int
 send_fd_to_socket (int server, int fd, libcrun_error_t *err)
+{
+  return send_fd_to_socket_with_payload (server, fd, NULL, 0, err);
+}
+
+int
+send_fd_to_socket_with_payload (int server, int fd, const char *payload, size_t payload_len, libcrun_error_t *err)
 {
   int ret;
   struct cmsghdr *cmsg = NULL;
-  struct iovec iov[1];
+  struct iovec iov[2];
   struct msghdr msg = {};
-  char ctrl_buf[CMSG_SPACE (sizeof (int))] = {};
+  char ctrl_buf[CMSG_SPACE (1 + sizeof (int))] = {};
   char data[1];
 
   data[0] = ' ';
   iov[0].iov_base = data;
   iov[0].iov_len = sizeof (data);
 
+  if (payload_len > 0)
+    {
+      iov[0].iov_base = (void *) payload;
+      iov[0].iov_len = payload_len;
+    }
+
   msg.msg_name = NULL;
   msg.msg_namelen = 0;
   msg.msg_iov = iov;
@@ -907,7 +1019,7 @@ send_fd_to_socket (int server, int fd, libcrun_error_t *err)
 }
 
 int
-receive_fd_from_socket (int from, libcrun_error_t *err)
+receive_fd_from_socket_with_payload (int from, char *payload, size_t payload_len, libcrun_error_t *err)
 {
   cleanup_close int fd = -1;
   int ret;
@@ -921,6 +1033,12 @@ receive_fd_from_socket (int from, libcrun_error_t *err)
   iov[0].iov_base = data;
   iov[0].iov_len = sizeof (data);
 
+  if (payload_len > 0)
+    {
+      iov[0].iov_base = (void *) payload;
+      iov[0].iov_len = payload_len;
+    }
+
   msg.msg_name = NULL;
   msg.msg_namelen = 0;
   msg.msg_iov = iov;
@@ -931,6 +1049,8 @@ receive_fd_from_socket (int from, libcrun_error_t *err)
   ret = TEMP_FAILURE_RETRY (recvmsg (from, &msg, 0));
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "recvmsg");
+  if (UNLIKELY (ret == 0))
+    return crun_make_error (err, 0, "read FD: connection closed");
 
   cmsg = CMSG_FIRSTHDR (&msg);
   if (cmsg == NULL)
@@ -942,6 +1062,12 @@ receive_fd_from_socket (int from, libcrun_error_t *err)
   return ret;
 }
 
+int
+receive_fd_from_socket (int from, libcrun_error_t *err)
+{
+  return receive_fd_from_socket_with_payload (from, NULL, 0, err);
+}
+
 int
 create_socket_pair (int *pair, libcrun_error_t *err)
 {
@@ -1036,8 +1162,7 @@ copy_from_fd_to_fd (int src, int dst, int consume, libcrun_error_t *err)
             return crun_make_error (err, errno, "write");
           remaining -= ret;
         }
-    }
-  while (consume && nread);
+  } while (consume && nread);
 
   return 0;
 }
@@ -1148,7 +1273,8 @@ set_home_env (uid_t id)
   if (stream == NULL)
     {
       if (errno == ENOENT)
-        return 0;
+        goto exit;
+
       return -1;
     }
 
@@ -1177,6 +1303,10 @@ set_home_env (uid_t id)
           return 0;
         }
     }
+
+exit:
+  /* If the user was not found, set it to something reasonable.  */
+  setenv ("HOME", "/", 1);
   return 0;
 }
 
@@ -1418,7 +1548,7 @@ mark_for_close_fds_ge_than (int n, libcrun_error_t *err)
   ret = syscall_close_range (n, UINT_MAX, CLOSE_RANGE_CLOEXEC);
   if (ret == 0)
     return 0;
-  if (ret < 0 && errno != EINVAL && errno != ENOSYS)
+  if (ret < 0 && errno != EINVAL && errno != ENOSYS && errno != EPERM)
     return crun_make_error (err, errno, "close_range from %d", n);
 
   cfd = open ("/proc/self/fd", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
@@ -1512,7 +1642,11 @@ check_access (const char *path)
   int ret;
   mode_t mode;
 
+#ifdef ANDROID
+  ret = access (path, X_OK);
+#else
   ret = eaccess (path, X_OK);
+#endif
   if (ret < 0)
     return ret;
 
@@ -1530,7 +1664,7 @@ check_access (const char *path)
 }
 
 const char *
-find_executable (const char *executable_path, const char *cwd)
+find_executable (const char *executable_path, const char *cwd, const char *handler)
 {
   cleanup_free char *cwd_executable_path = NULL;
   cleanup_free char *tmp = NULL;
@@ -1551,7 +1685,7 @@ find_executable (const char *executable_path, const char *cwd)
 
       if (cwd == NULL)
         {
-          cwd_allocated = get_current_dir_name ();
+          cwd_allocated = getcwd (NULL, 0);
           if (cwd_allocated == NULL)
             OOM ();
 
@@ -1564,6 +1698,11 @@ find_executable (const char *executable_path, const char *cwd)
       executable_path = cwd_executable_path;
     }
 
+  /* Process executable according to configured handler */
+  /* For `wasm` we don't want to check access hence return early*/
+  if (handler != NULL && (strcmp (handler, "wasm") == 0))
+    return xstrdup (executable_path);
+
   /* Absolute path.  It doesn't need to lookup $PATH.  */
   if (executable_path[0] == '/')
     {
@@ -1693,7 +1832,9 @@ copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size, dev
   int ret;
 
 #ifdef HAVE_STATX
-  struct statx stx;
+  struct statx stx = {
+    0,
+  };
 
   ret = statx (dfd, path, AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC,
                STATX_TYPE | STATX_MODE | STATX_SIZE | STATX_UID | STATX_GID, &stx);
@@ -1745,8 +1886,6 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
       cleanup_close int srcfd = -1;
       cleanup_close int destfd = -1;
       cleanup_free char *target_buf = NULL;
-      ssize_t buf_size;
-      ssize_t size;
       int ret;
       mode_t mode;
       off_t st_size;
@@ -1812,20 +1951,9 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
           break;
 
         case S_IFLNK:
-          buf_size = st_size + 1;
-          target_buf = xmalloc (buf_size);
-
-          do
-            {
-              buf_size += 1024;
-
-              target_buf = xrealloc (target_buf, buf_size);
-
-              size = readlinkat (dirfd (dsrcfd), de->d_name, target_buf, buf_size);
-              if (UNLIKELY (size < 0))
-                return crun_make_error (err, errno, "readlink `%s/%s`", srcname, de->d_name);
-            }
-          while (size == buf_size);
+          ret = safe_readlinkat (dirfd (dsrcfd), de->d_name, &target_buf, st_size, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
 
           ret = symlinkat (target_buf, destdirfd, de->d_name);
           if (UNLIKELY (ret < 0))
@@ -1984,7 +2112,7 @@ append_paths (char **out, libcrun_error_t *err, ...)
       bool has_trailing_slash;
 
       has_trailing_slash = copied > 0 && (*out)[copied - 1] == '/';
-      if (i > 0 && !has_trailing_slash)
+      if (i > 0 && ! has_trailing_slash)
         {
           (*out)[copied] = '/';
           copied += 1;
@@ -1996,3 +2124,109 @@ append_paths (char **out, libcrun_error_t *err, ...)
   (*out)[copied] = '\0';
   return 0;
 }
+
+/* Adapted from mailutils 0.6.91 (distributed under LGPL 2.0+)  */
+static int
+b64_input (char c)
+{
+  const char table[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+  int i;
+
+  for (i = 0; i < 64; i++)
+    {
+      if (table[i] == c)
+        return i;
+    }
+  return -1;
+}
+
+int
+base64_decode (const char *iptr, size_t isize, char *optr, size_t osize, size_t *nbytes)
+{
+  int i = 0, tmp = 0, pad = 0;
+  size_t consumed = 0;
+  unsigned char data[4];
+
+  *nbytes = 0;
+  while (consumed < isize && (*nbytes) + 3 < osize)
+    {
+      while ((i < 4) && (consumed < isize))
+        {
+          tmp = b64_input (*iptr++);
+          consumed++;
+          if (tmp != -1)
+            data[i++] = tmp;
+          else if (*(iptr - 1) == '=')
+            {
+              data[i++] = '\0';
+              pad++;
+            }
+        }
+
+      /* I have a entire block of data 32 bits get the output data.  */
+      if (i == 4)
+        {
+          *optr++ = (data[0] << 2) | ((data[1] & 0x30) >> 4);
+          *optr++ = ((data[1] & 0xf) << 4) | ((data[2] & 0x3c) >> 2);
+          *optr++ = ((data[2] & 0x3) << 6) | data[3];
+          (*nbytes) += 3 - pad;
+        }
+      else
+        {
+          /* I did not get all the data.  */
+          consumed -= i;
+          return consumed;
+        }
+      i = 0;
+    }
+  return consumed;
+}
+
+char *
+get_user_name (uid_t uid)
+{
+  struct passwd pd;
+  struct passwd *temp_result_ptr;
+  char pwdbuffer[200];
+  if (! getpwuid_r (uid, &pd, pwdbuffer, sizeof (pwdbuffer), &temp_result_ptr))
+    return xstrdup (pd.pw_name);
+  return xstrdup ("");
+}
+
+int
+has_suffix (const char *str, const char *suffix)
+{
+  if (! str || ! suffix)
+    return 0;
+  size_t lenstr = strlen (str);
+  size_t lensuffix = strlen (suffix);
+  if (lensuffix > lenstr)
+    return 0;
+  return memcmp (str + lenstr - lensuffix, suffix, lensuffix) == 0;
+}
+
+char *
+str_join_array (int offset, size_t size, char *const array[], const char *joint)
+{
+  size_t jlen, lens[size];
+  size_t i, total_size = (size - 1) * (jlen = strlen (joint)) + 1;
+  char *result, *p;
+
+  for (i = 0; i < size; ++i)
+    {
+      total_size += (lens[i] = strlen (array[i]));
+    }
+  p = result = xmalloc (total_size);
+  for (i = offset; i < size; ++i)
+    {
+      memcpy (p, array[i], lens[i]);
+      p += lens[i];
+      if (i < size - 1)
+        {
+          memcpy (p, joint, jlen);
+          p += jlen;
+        }
+    }
+  *p = '\0';
+  return result;
+}
