@@ -26,7 +26,9 @@
 #include "error.h"
 #include <dirent.h>
 #include <unistd.h>
+#include <signal.h>
 #include <runtime_spec_schema_config_schema.h>
+#include <sys/wait.h>
 #include "container.h"
 
 #ifndef TEMP_FAILURE_RETRY
@@ -44,8 +46,10 @@
 #define cleanup_free __attribute__ ((cleanup (cleanup_freep)))
 #define cleanup_close __attribute__ ((cleanup (cleanup_closep)))
 #define cleanup_close_vec __attribute__ ((cleanup (cleanup_close_vecp)))
+#define cleanup_close_map __attribute__ ((cleanup (cleanup_close_mapp)))
 #define cleanup_dir __attribute__ ((cleanup (cleanup_dirp)))
 #define arg_unused __attribute__ ((unused))
+#define cleanup_pid __attribute__ ((cleanup (cleanup_pidp)))
 
 #define LIKELY(x) __builtin_expect ((x), 1)
 #define UNLIKELY(x) __builtin_expect ((x), 0)
@@ -100,6 +104,54 @@ cleanup_closep (void *p)
     TEMP_FAILURE_RETRY (close (*pp));
 }
 
+static inline void
+cleanup_pidp (void *p)
+{
+  pid_t *pp = p;
+  if (*pp > 0)
+    {
+      TEMP_FAILURE_RETRY (kill (*pp, SIGKILL));
+      TEMP_FAILURE_RETRY (waitpid (*pp, NULL, 0));
+    }
+}
+
+struct
+    libcrun_fd_map
+{
+  size_t nfds;
+  int fds[];
+};
+
+static inline struct libcrun_fd_map *
+make_libcrun_fd_map (size_t len)
+{
+  struct libcrun_fd_map *ret;
+  size_t i;
+
+  ret = xmalloc (sizeof (*ret) + sizeof (int) * len);
+  ret->nfds = len;
+  for (i = 0; i < len; i++)
+    ret->fds[i] = -1;
+
+  return ret;
+}
+
+static inline void
+cleanup_close_mapp (struct libcrun_fd_map **p)
+{
+  struct libcrun_fd_map *m = *p;
+  size_t i;
+
+  if (m == NULL)
+    return;
+
+  for (i = 0; i < m->nfds; i++)
+    if (m->fds[i] >= 0)
+      TEMP_FAILURE_RETRY (close (m->fds[i]));
+
+  free (m);
+}
+
 static inline void
 cleanup_close_vecp (int **p)
 {
@@ -145,6 +197,14 @@ xstrdup (const char *str)
   return ret;
 }
 
+static inline const char *
+consume_slashes (const char *t)
+{
+  while (*t == '/')
+    t++;
+  return t;
+}
+
 int xasprintf (char **str, const char *fmt, ...);
 
 int crun_path_exists (const char *path, libcrun_error_t *err);
@@ -163,6 +223,8 @@ int crun_ensure_directory_at (int dirfd, const char *path, int mode, bool nofoll
 
 int crun_ensure_file_at (int dirfd, const char *path, int mode, bool nofollow, libcrun_error_t *err);
 
+int crun_safe_create_and_open_ref_at (bool dir, int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode, libcrun_error_t *err);
+
 int crun_safe_ensure_directory_at (int dirfd, const char *dirpath, size_t dirpath_len, const char *path, int mode,
                                    libcrun_error_t *err);
 
@@ -197,10 +259,14 @@ int open_unix_domain_socket (const char *path, int dgram, libcrun_error_t *err);
 
 int send_fd_to_socket (int server, int fd, libcrun_error_t *err);
 
+int send_fd_to_socket_with_payload (int server, int fd, const char *payload, size_t payload_len, libcrun_error_t *err);
+
 int create_socket_pair (int *pair, libcrun_error_t *err);
 
 int receive_fd_from_socket (int from, libcrun_error_t *err);
 
+int receive_fd_from_socket_with_payload (int from, char *payload, size_t payload_len, libcrun_error_t *err);
+
 int create_signalfd (sigset_t *mask, libcrun_error_t *err);
 
 int epoll_helper (int *fds, int *levelfds, libcrun_error_t *err);
@@ -224,7 +290,7 @@ int parse_json_file (yajl_val *out, const char *jsondata, struct parser_context
 
 int has_prefix (const char *str, const char *prefix);
 
-const char *find_executable (const char *executable_path, const char *cwd);
+const char *find_executable (const char *executable_path, const char *cwd, const char *handler);
 
 int copy_recursive_fd_to_fd (int srcfd, int destfd, const char *srcname, const char *destname, libcrun_error_t *err);
 
@@ -242,6 +308,8 @@ int get_file_type (mode_t *mode, bool nofollow, const char *path);
 
 int get_file_type_fd (int fd, mode_t *mode);
 
+char *get_user_name (uid_t uid);
+
 int safe_openat (int dirfd, const char *rootfs, size_t rootfs_len, const char *path, int flags, int mode,
                  libcrun_error_t *err);
 
@@ -251,4 +319,8 @@ int append_paths (char **out, libcrun_error_t *err, ...);
 
 LIBCRUN_PUBLIC int libcrun_str2sig (const char *name);
 
+int base64_decode (const char *iptr, size_t isize, char *optr, size_t osize, size_t *nbytes);
+int has_suffix (const char *source, const char *suffix);
+char *str_join_array (int offset, size_t size, char *const array[], const char *joint);
+
 #endif
