@@ -18,19 +18,20 @@
 
 #include <libcrun/error.h>
 #include <libcrun/utils.h>
+#include <libcrun/cgroup.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <string.h>
 
-typedef int (*test)();
+typedef int (*test) ();
 
 static int
 test_socket_pair ()
 {
   libcrun_error_t err = NULL;
   int fds[2];
-  cleanup_close int fd0 = -1;
-  cleanup_close int fd1 = -1;
+  __attribute__ ((unused)) cleanup_close int fd0 = -1;
+  __attribute__ ((unused)) cleanup_close int fd1 = -1;
   char buffer[256];
   int ret = create_socket_pair (fds, &err);
   if (ret < 0)
@@ -95,7 +96,7 @@ test_send_receive_fd ()
       char buffer[256];
       const char *test_string = "TEST STRING";
       if (pipe (pipes) < 0)
-       return -1;
+        return -1;
 
       pipefd0 = pipes[0];
       pipefd1 = pipes[1];
@@ -132,7 +133,7 @@ test_send_receive_fd ()
       if (ret <= 0)
         return -1;
       if (write (fd0, buffer, ret) < 0)
-       return -1;
+        return -1;
 
       _exit (0);
     }
@@ -144,13 +145,13 @@ test_run_process ()
 {
   libcrun_error_t err = NULL;
   {
-    char *args[] = {"/bin/true", NULL};
+    char *args[] = { "/bin/true", NULL };
     if (run_process (args, &err) != 0)
       return -1;
   }
 
   {
-    char *args[] = {"/bin/false", NULL};
+    char *args[] = { "/bin/false", NULL };
     if (run_process (args, &err) <= 0)
       return -1;
     if (err != NULL)
@@ -158,7 +159,7 @@ test_run_process ()
   }
 
   {
-    char *args[] = {"/does/not/exist", NULL};
+    char *args[] = { "/does/not/exist", NULL };
     int r = run_process (args, &err);
     if (r <= 0)
       return -1;
@@ -250,20 +251,20 @@ test_crun_path_exists ()
 static int
 test_append_paths ()
 {
-#define PROLOGUE()                              \
-  cleanup_free char *out = NULL;                \
-  libcrun_error_t err = NULL;                   \
+#define PROLOGUE()               \
+  cleanup_free char *out = NULL; \
+  libcrun_error_t err = NULL;    \
   int ret;
 
-#define EXPECT_STRING(exp)                      \
-  {                                             \
-    if (ret < 0 || out == NULL)                 \
-      {                                         \
-        crun_error_release (&err);              \
-        return ret;                             \
-      }                                         \
-    if (strcmp (out, exp))                      \
-      return -1;                                \
+#define EXPECT_STRING(exp)         \
+  {                                \
+    if (ret < 0 || out == NULL)    \
+      {                            \
+        crun_error_release (&err); \
+        return ret;                \
+      }                            \
+    if (strcmp (out, exp))         \
+      return -1;                   \
   }
   {
     PROLOGUE ();
@@ -322,6 +323,52 @@ test_append_paths ()
   return 0;
 }
 
+#ifdef HAVE_SYSTEMD
+static int
+test_parse_sd_array ()
+{
+  char *out, *next;
+  libcrun_error_t err = NULL;
+  {
+    cleanup_free char *s = xstrdup ("       'firewalld.service'    ");
+    if (parse_sd_array (s, &out, &next, &err) < 0)
+      {
+        crun_error_release (&err);
+        return -1;
+      }
+    if (strcmp (out, "firewalld.service"))
+      return -1;
+    if (next != NULL)
+      return -1;
+  }
+  {
+    cleanup_free char *s = xstrdup ("'fir\\ewalld.ser\\vi\\ce']");
+    if (parse_sd_array (s, &out, &next, &err) < 0)
+      {
+        crun_error_release (&err);
+        return -1;
+      }
+    if (strcmp (out, "firewalld.service"))
+      return -1;
+    if (next != NULL)
+      return -1;
+  }
+  {
+    cleanup_free char *s = xstrdup ("'fi\\rew\\alld.s\\erv\\ice', 'foo.service']");
+    if (parse_sd_array (s, &out, &next, &err) < 0)
+      {
+        crun_error_release (&err);
+        return -1;
+      }
+    if (strcmp (out, "firewalld.service"))
+      return -1;
+    if (strcmp (next, " 'foo.service']"))
+      return -1;
+  }
+  return 0;
+}
+#endif
+
 static void
 run_and_print_test_result (const char *name, int id, test t)
 {
@@ -334,13 +381,21 @@ run_and_print_test_result (const char *name, int id, test t)
     printf ("not ok %d - %s\n", id, name);
 }
 
-#define RUN_TEST(T) do {run_and_print_test_result (#T, id++, T);} while (0)
+#define RUN_TEST(T)                            \
+  do                                           \
+    {                                          \
+      run_and_print_test_result (#T, id++, T); \
+  } while (0)
 
 int
 main ()
 {
   int id = 1;
+#ifdef HAVE_SYSTEMD
+  printf ("1..8\n");
+#else
   printf ("1..7\n");
+#endif
   RUN_TEST (test_crun_path_exists);
   RUN_TEST (test_write_read_file);
   RUN_TEST (test_run_process);
@@ -348,5 +403,8 @@ main ()
   RUN_TEST (test_socket_pair);
   RUN_TEST (test_send_receive_fd);
   RUN_TEST (test_append_paths);
+#ifdef HAVE_SYSTEMD
+  RUN_TEST (test_parse_sd_array);
+#endif
   return 0;
 }
